import logging
import os
import tempfile

import falcon

from galileo.apps.repository import Repository, RepositoryException

logger = logging.getLogger(__name__)


class RepositoryResource:
    repo: Repository

    def __init__(self, repo: Repository) -> None:
        super().__init__()
        self.repo = repo

    def on_get(self, req, resp):
        resp.media = [{'name': app.name, 'manifest': app.manifest} for app in self.repo.list_apps()]

    def on_get_info(self, req, resp, app_name):
        app = self.repo.get_app(app_name)
        if not app:
            raise falcon.HTTPNotFound()

        resp.media = {'name': app.name, 'manifest': app.manifest}

    def on_delete_info(self, req, resp, app_name):
        resp.media = self.repo.delete_app(app_name)

    def on_post(self, req: falcon.Request, resp: falcon.Response):
        with tempfile.TemporaryDirectory() as tmpdir:
            zip_path = os.path.join(tmpdir, 'app.zip')

            with open(zip_path, 'wb') as fd:
                fd.write(req.stream.read())

            try:
                app = self.repo.add(zip_path)
            except RepositoryException as e:
                raise falcon.HTTPBadRequest('Error processing upload', str(e))

        resp.status = falcon.HTTP_201
        resp.media = {'name': app.name, 'manifest': app.manifest}

    def on_get_download(self, req, resp: falcon.Response, app_name):
        app = self.repo.get_app(app_name)
        if not app:
            raise falcon.HTTPNotFound()

        resp.content_type = 'application/zip'
        with open(app.archive_path, 'rb') as fd:
            resp.body = fd.read()

        resp.downloadable_as = app.archive_path.split('/')[-1]



import abc
import logging
import time
import requests

from galileo.routing.balancer import Balancer

logger = logging.getLogger(__name__)


class ServiceRequest:
    service: str
    path: str
    method: str
    kwargs: dict

    created: float
    sent: float
    done: float

    def __init__(self, service, path='/', method='get', **kwargs) -> None:
        super().__init__()
        self.service = service
        self.path = path
        self.method = method
        self.kwargs = kwargs

        self.created = time.time()


class Router(abc.ABC):

    def __init__(self):
        self.last_log_update = time.time()
        self.requests_since_last_log_update = 0

    def request(self, req: ServiceRequest) -> requests.Response:
        url = self._get_url(req)

        logger.debug('forwarding request %s %s', req.method, url)

        req.sent = time.time()
        response = requests.request(req.method, url, **req.kwargs)
        req.done = req.sent

        logger.debug('%s %s: %s', req.method, url, response.status_code)
        self.requests_since_last_log_update += 1
        if time.time() - self.last_log_update >= 1:
            logger.debug(f'Sent {self.requests_since_last_log_update} requests in the last second')
            self.requests_since_last_log_update = 0
            self.last_log_update = time.time()
        return response

    def _get_url(self, req: ServiceRequest) -> str:
        raise NotImplementedError


class StaticRouter(Router):
    """
    Routes a request statically to <path_prefix>/<request_path>, so StaticRouter('http://localhost:8080/') would route
    everything to a webserver on localhost:8080.
    """

    def __init__(self, path_prefix) -> None:
        super().__init__()
        self.path_prefix = path_prefix

    def _get_url(self, req: ServiceRequest) -> str:
        return f'{self.path_prefix}{req.path}'


class DynamicRouter(Router, abc.ABC):
    """
    Abstract base class for routing using a balancer.
    """
    _balancer: Balancer

    def __init__(self, balancer: Balancer) -> None:
        super().__init__()
        self._balancer = balancer

    def _get_url(self, req: ServiceRequest) -> str:
        host = self._balancer.next_host(req.service)
        return self._create_url(host, req)

    def _create_url(self, host, req: ServiceRequest):
        raise NotImplementedError


class HostRouter(DynamicRouter):
    """
    Routes a service request to http://<host>/<request_path> using a balancer. This is useful for testing without
    requiring a symmetry proxy on each mc2 node for routing to the correct container.
    """

    def _create_url(self, host, req: ServiceRequest):
        return f'http://{host}{req.path}'


class ServiceRouter(DynamicRouter):
    """
    Routes a service request to http://<host>/<service>/<request_path> using a balancer. This assumes a setup where each
    mc2 node has a reverse proxy that routes requests to the correct app container.
    """

    def _create_url(self, host, req: ServiceRequest) -> str:
        return f'http://{host}/{req.service}{req.path}'



import logging

import pymq
import redis
from galileodb import Experiment
from galileodb.model import QueuedExperiment, ExperimentConfiguration
from pymq.provider.redis import RedisQueue
from redis import WatchError

from galileo.controller import RedisClusterController

logger = logging.getLogger(__name__)


class CancelError(Exception):
    pass


class ExperimentQueue:
    queue: RedisQueue

    def __init__(self, queue) -> None:
        super().__init__()
        if not isinstance(queue, RedisQueue):
            raise ValueError

        self.queue = queue

    def __getattr__(self, item):
        if hasattr(self.queue, item):
            return getattr(self.queue, item)
        raise AttributeError

    def remove(self, experiment_id, retries=3):
        queue = self.queue
        key = queue._key
        rds = queue._rds

        for i in range(retries):
            try:
                with rds.pipeline() as pipe:
                    pipe.watch(key)
                    workloads = pipe.lrange(key, 0, -1)
                    pipe.multi()
                    index = -1
                    for idx, entry in enumerate(workloads):
                        item: QueuedExperiment = queue._deserialize(entry)
                        if item.experiment.id == experiment_id:
                            index = idx
                            break

                    if index == -1:
                        return False

                    pipe.lset(key, index, 'DELETE')
                    pipe.lrem(key, 1, 'DELETE')
                    pipe.execute()
                    return True
            except WatchError:
                logger.warning('WatchError cancelling experiment with id %s (try %d)', experiment_id, i)
                continue

        raise CancelError


class ExperimentController:
    queue_key = 'galileo:experiments:queue'

    def __init__(self, rds: redis.Redis = None, eventbus=None) -> None:
        super().__init__()
        self.rds = rds or redis.Redis(decode_responses=True)
        self.eventbus = eventbus or pymq
        self.cluster = RedisClusterController(self.rds, eventbus=self.eventbus)
        self.experiment_queue = ExperimentQueue(self.eventbus.queue(self.queue_key))

    def queue(self, config: ExperimentConfiguration, exp: Experiment = None):
        """
        Queues an experiment for the experiment daemon to load.
        :param config: experiment configuration
        :param exp: the experiment metadata (optional, as all parameters could be generated)
        :return:
        """
        if not self.cluster.list_workers():
            raise ValueError('No workers to execute the experiment on')

        element = QueuedExperiment(exp, config)
        logger.debug('queuing experiment data: %s', element)
        self.experiment_queue.put(element)

    def cancel(self, exp_id: str) -> bool:
        return self.experiment_queue.remove(exp_id)



import logging
import threading
import unittest
from unittest.mock import patch

import pymq
from galileodb import ExperimentDatabase
from galileodb.model import QueuedExperiment, Experiment, ExperimentConfiguration, WorkloadConfiguration
from galileodb.recorder import ExperimentTelemetryRecorder
from pymq.provider.redis import RedisConfig
from timeout_decorator import timeout_decorator

from galileo.controller import ExperimentController, RedisClusterController
from galileo.experiment.experimentd import ExperimentDaemon
from galileo.experiment.service import SimpleExperimentService
from galileo.util import poll
from tests.testutils import RedisResource, SqliteResource

logging.basicConfig(level=logging.DEBUG)


@unittest.skip  # FIXME
class TestExperimentDaemon(unittest.TestCase):
    exp_db: ExperimentDatabase
    redis_resource: RedisResource = RedisResource()

    def setUp(self) -> None:
        self.rds = self.init_rds()
        self.exp_db = self.init_db()
        pymq.init(RedisConfig(self.rds))

        self.recorder_factory = lambda exp_id: ExperimentTelemetryRecorder(self.rds, self.exp_db, exp_id)
        self.exp_ctrl = ExperimentController(self.rds)
        self.cctrl = RedisClusterController(self.rds)
        self.exp_service = SimpleExperimentService(self.exp_db)

    def tearDown(self) -> None:
        pymq.shutdown()
        self.redis_resource.tearDown()
        self.db_resource.tearDown()

    @patch('galileo.experiment.runner.run_experiment')
    @timeout_decorator.timeout(30)
    def test_integration(self, mocked_run_experiment):
        self.cctrl.register_worker('host1')  # create a worker

        daemon = ExperimentDaemon(self.rds, self.recorder_factory, self.exp_ctrl, self.exp_service)

        def inject_experiment():
            exp = Experiment('experiment_id', creator='unittest')
            cfg = ExperimentConfiguration(2, 1, [WorkloadConfiguration('aservice', [3, 5], 2, 'constant')])

            queue = pymq.queue(ExperimentController.queue_key)
            queue.put(QueuedExperiment(exp, cfg))
            try:
                poll(lambda: queue.qsize() == 0, timeout=2, interval=0.1)  # wait for the daemon to take the experiment
            finally:
                daemon.close()

        threading.Thread(target=inject_experiment).start()
        daemon.run()

        # wait for the experiment to be created
        poll(lambda: self.exp_service.exists('experiment_id'), timeout=2, interval=0.1)
        # wait for the experiment to be finished
        poll(lambda: self.exp_service.find('experiment_id').status == 'FINISHED', timeout=3, interval=0.1)

        # verify that the experiment parameters were set correctly
        exp = self.exp_service.find('experiment_id')
        self.assertIsNotNone(exp)
        self.assertEqual('FINISHED', exp.status)
        self.assertEqual('experiment_id', exp.id)
        self.assertEqual('experiment_id', exp.name)

        # verify that experiment daemon tried to run the experiment
        self.assertTrue(mocked_run_experiment.called, 'expected ExperimentDaemon to use experiment runner')

    def init_rds(self):
        self.redis_resource.setUp()
        return self.redis_resource.rds

    def init_db(self):
        self.db_resource = SqliteResource()
        self.db_resource.setUp()
        return self.db_resource.db


if __name__ == '__main__':
    unittest.main()



from typing import NamedTuple


class RegisterWorkerEvent(NamedTuple):
    name: str


class UnregisterWorkerEvent(NamedTuple):
    name: str


class RegisterWorkerCommand(NamedTuple):
    pass


class StartTracingCommand(NamedTuple):
    pass


class PauseTracingCommand(NamedTuple):
    pass


class ClientConfig(NamedTuple):
    service: str
    client: str = None
    parameters: dict = None
    worker_labels: dict = None

    def __repr__(self):
        return self.__str__()

    def __str__(self) -> str:
        return 'ClientConfig(service={}, client={}, parameters={}, worker_labels:{})'.format(
            self.service, self.client, self._abbrv_parameters(self.parameters),
            self.worker_labels)

    def _abbrv_parameters(self, d):
        if d is None:
            return d

        s = str(d)
        if len(s) > 120:
            return s[:105] + ' <abbreviated> }'


class ClientDescription(NamedTuple):
    client_id: str
    worker: str
    config: ClientConfig


class ClientInfo(NamedTuple):
    description: ClientDescription
    requests: int
    failed: int


class CloseClientCommand(NamedTuple):
    client_id: str


class CreateClientCommand(NamedTuple):
    host: str
    config: ClientConfig
    num: int = 1


class StopClientsCommand(NamedTuple):
    client_id: str


class ClientStartedEvent(NamedTuple):
    host: str
    client_id: str


class ClientStoppedEvent(NamedTuple):
    host: str
    client_id: str


class SetWorkloadCommand(NamedTuple):
    client_id: str
    num: int = None
    distribution: str = 'constant'
    parameters: tuple = None


class StopWorkloadCommand(NamedTuple):
    client_id: str


class WorkloadDoneEvent(NamedTuple):
    client_id: str



import logging


def on_starting(server):
    # simple way of getting the logs to output by reusing gunicorn's --log-level config
    logging.basicConfig(level=logging.getLogger('gunicorn.error').level)


def on_exit(server):
    pass



import os
import threading
import time
import unittest
from queue import Queue
from typing import List, NamedTuple
from unittest.mock import patch

from pymq.provider.simple import SimpleEventBus
from timeout_decorator import timeout_decorator

from galileo.routing import ServiceRequest, RedisRoutingTable, RoutingRecord
from galileo.worker.api import ClientDescription, ClientConfig, SetWorkloadCommand
from galileo.worker.client import Client, RequestGenerator, single_request
from galileo.worker.context import Context, DebugRouter
from tests.testutils import RedisResource


class StaticRequestGenerator:

    def __init__(self, requests: List[ServiceRequest]) -> None:
        super().__init__()
        self.requests = requests

    def run(self):
        yield from self.requests

    def close(self):
        pass

    def set_rps(self, *args, **kwargs):
        pass


class ClientTest(unittest.TestCase):

    @timeout_decorator.timeout(5)
    def test_client_integration(self):
        env = dict(os.environ)
        env['galileo_router_type'] = 'DebugRouter'

        client_id = 'unittest_client'
        ctx = Context(env)
        trace_queue = Queue()

        description = ClientDescription(client_id, 'unittest_worker', ClientConfig('aservice'))
        # ctx: Context, trace_queue: Queue, description: ClientDescription

        client = Client(ctx, trace_queue, description, eventbus=SimpleEventBus())

        client.request_generator = StaticRequestGenerator([
            ServiceRequest('aservice'),
            ServiceRequest('aservice'),
        ])

        client.run()

        trace1 = trace_queue.get(timeout=2)
        trace2 = trace_queue.get(timeout=2)

        self.assertEqual('aservice', trace1.service)
        self.assertEqual('aservice', trace2.service)

        self.assertEqual('debughost', trace1.server)
        self.assertEqual('debughost', trace2.server)

        now = time.time()
        self.assertAlmostEqual(now, trace1.done, delta=2)
        self.assertAlmostEqual(now, trace2.done, delta=2)

    @timeout_decorator.timeout(5)
    def test_with_router_fault(self):
        class FaultInjectingRouter(DebugRouter):
            def request(self, req: ServiceRequest) -> 'requests.Response':
                if req.path == '/api/nonexisting':
                    raise ValueError('some error')

                return super().request(req)

        router = FaultInjectingRouter()

        ctx = Context()
        ctx.create_router = lambda: router
        client_id = 'unittest_client'
        trace_queue = Queue()

        description = ClientDescription(client_id, 'unittest_worker', ClientConfig('aservice'))
        # ctx: Context, trace_queue: Queue, description: ClientDescription

        client = Client(ctx, trace_queue, description, eventbus=SimpleEventBus())

        client.request_generator = StaticRequestGenerator([
            ServiceRequest('aservice', path='/api/nonexisting'),
            ServiceRequest('aservice', path='/api/unittest'),
        ])

        client.run()

        trace1 = trace_queue.get(timeout=2)
        trace2 = trace_queue.get(timeout=2)

        self.assertEqual(-1, trace1.sent)
        self.assertAlmostEqual(trace2.sent, time.time(), delta=2)


class TestSingleRequest(unittest.TestCase):
    redis_resource = RedisResource()

    def setUp(self) -> None:
        self.redis_resource.setUp()
        self.rds = self.redis_resource.rds
        self.rtbl = RedisRoutingTable(self.rds)

    def tearDown(self) -> None:
        self.redis_resource.tearDown()

    @patch('galileo.routing.router.requests.request')
    def test_single_request(self, mock_request):
        # mock http server
        Response = NamedTuple('Response', status_code=int, text=str, url=str, method=str)

        def fake_response(method, url):
            return Response(200, 'ok', url, method)

        mock_request.side_effect = fake_response

        # overwrite `Context` used by client
        ctx = Context()
        ctx.create_redis = lambda: self.redis_resource.rds

        # create routing record
        self.rtbl.set_routing(RoutingRecord('myservice', ['localhost:8000'], [1]))

        # make request
        response = single_request(ClientConfig('myservice'), ctx, router_type='SymmetryHostRouter')

        self.assertEqual(200, response.status_code)


def queue_collect(vq: Queue, gen):
    for v in gen:
        vq.put(v)


class RequestGeneratorTest(unittest.TestCase):
    def test_with_limit_and_no_interval(self):
        workload = SetWorkloadCommand('myclient', 3)
        request_generator = RequestGenerator(lambda: 1)
        q = Queue()

        t = threading.Thread(target=queue_collect, args=(q, request_generator.run()))
        t.start()

        try:
            request_generator.set_workload(workload)
            self.assertEqual(1, q.get(timeout=1))
            self.assertEqual(1, q.get(timeout=1))
            self.assertEqual(1, q.get(timeout=1))
            self.assertEqual(RequestGenerator.DONE, q.get())
            self.assertEqual(0, q.qsize())
        finally:
            request_generator.close()
            t.join(2)

    def test_with_interval(self):
        workload = SetWorkloadCommand('myclient', parameters=(0.1,))  # 0.1 interarrival delay
        request_generator = RequestGenerator(lambda: 1)
        q = Queue()

        t = threading.Thread(target=queue_collect, args=(q, request_generator.run()))
        t.start()

        try:
            request_generator.set_workload(workload)
            time.sleep(2)
            self.assertAlmostEqual(20, q.qsize(), delta=3)
        finally:
            request_generator.close()
            t.join(2)

    def test_with_interval_and_limit(self):
        workload = SetWorkloadCommand('myclient', num=10, parameters=(0.05,))  # 0.05 interarrival delay
        request_generator = RequestGenerator(lambda: 1)
        q = Queue()

        t = threading.Thread(target=queue_collect, args=(q, request_generator.run()))
        t.start()

        try:
            request_generator.set_workload(workload)

            while True:
                item = q.get(timeout=2)
                if item is RequestGenerator.DONE:
                    break

            self.assertEqual(0, q.qsize())
        finally:
            request_generator.close()
            t.join(2)



import multiprocessing
import queue
import threading
import unittest
from unittest.mock import patch

import pymq
import redis
from galileodb.trace import START, PAUSE
from pymq.provider.redis import RedisConfig
from timeout_decorator import timeout_decorator

from galileo.worker.api import RegisterWorkerEvent, StartTracingCommand, PauseTracingCommand
from galileo.worker.context import Context
from galileo.worker.daemon import WorkerDaemon
from tests.testutils import RedisResource


class MockQueue:

    def __init__(self):
        super().__init__()
        self.traces = queue.Queue()
        self.delegate = multiprocessing.Queue()

    def put(self, *args, **kwargs) -> None:
        self.traces.put(args[0])
        self.delegate.put(*args, **kwargs)

    def __getattr__(self, item):
        if hasattr(self.delegate, item):
            return getattr(self.delegate, item)
        raise AttributeError


class WorkerTest(unittest.TestCase):
    redis_resource = RedisResource()
    eventbus: pymq.EventBus

    def setUp(self) -> None:
        self.redis_resource.setUp()
        self.eventbus = pymq.init(RedisConfig(self.redis_resource.rds))

    def tearDown(self) -> None:
        pymq.shutdown()
        self.redis_resource.tearDown()

    @timeout_decorator.timeout(5)
    def test_worker_start_register(self):
        rds = self.redis_resource.rds

        class TestContext(Context):
            def create_redis(self) -> redis.Redis:
                return rds

        register_event = threading.Event()

        def register_listener(event: RegisterWorkerEvent):
            register_event.set()

        self.eventbus.subscribe(register_listener)

        ctx = TestContext()

        worker = WorkerDaemon(ctx, eventbus=self.eventbus)

        worker_thread = threading.Thread(target=worker.run)
        worker_thread.start()

        self.assertTrue(register_event.wait(2), 'did not receive register event')

        worker.close()

        worker_thread.join()

    @timeout_decorator.timeout(5)
    def test_worker_start_logger(self):
        self.assert_msg_in_queue_after_cmd(StartTracingCommand(), START)

    @timeout_decorator.timeout(5)
    def test_worker_pause_logger(self):
        self.assert_msg_in_queue_after_cmd(PauseTracingCommand(), PAUSE)

    def assert_msg_in_queue_after_cmd(self, cmd, msg):
        trace_queue = MockQueue()

        started = threading.Event()

        def started_listener(event: RegisterWorkerEvent):
            started.set()

        self.eventbus.subscribe(started_listener)

        with patch.object(WorkerDaemon, '_create_trace_queue', return_value=trace_queue):
            rds = self.redis_resource.rds

            class TestContext(Context):
                def create_redis(self) -> redis.Redis:
                    return rds

            ctx = TestContext()

            worker = WorkerDaemon(ctx, eventbus=self.eventbus)

            worker_thread = threading.Thread(target=worker.run)
            worker_thread.start()

            started.wait(3)

            self.eventbus.publish(cmd)

            got = trace_queue.traces.get()
            self.assertEqual(msg, got)
            worker.close()
            worker_thread.join()



import time
from abc import ABC
from typing import List

from galileodb import Experiment, ExperimentDatabase


class ExperimentService(ABC):

    def save(self, exp: Experiment) -> bool:
        raise NotImplementedError

    def find(self, exp_id: str):
        raise NotImplementedError

    def delete(self, exp_id: str):
        raise NotImplementedError

    def finalize_experiment(self, exp: Experiment, status):
        raise NotImplementedError

    def exists(self, exp_id: str) -> bool:
        raise NotImplementedError

    def find_all(self) -> List[Experiment]:
        raise NotImplementedError


class SimpleExperimentService(ExperimentService):

    def __init__(self, repository: ExperimentDatabase):
        self.repository = repository

    def save(self, exp: Experiment) -> bool:
        if not self.exists(exp.id):
            self.repository.save_experiment(exp)
            return True
        else:
            self.repository.update_experiment(exp)
            return False

    def find(self, exp_id: str):
        return self.repository.get_experiment(exp_id)

    def delete(self, exp_id: str):
        self.repository.delete_experiment(exp_id)

    def finalize_experiment(self, exp: Experiment, status):
        exp.status = status
        exp.end = time.time()
        self.repository.update_experiment(exp)
        self.repository.touch_traces(exp)

    def exists(self, exp_id) -> bool:
        return self.repository.get_experiment(exp_id) is not None

    def find_all(self):
        return self.repository.find_all()



import unittest
import unittest.mock
from collections import Counter, defaultdict

from galileo.routing.balancer import WeightedRandomBalancer, WeightedRoundRobinBalancer
from galileo.routing.table import RoutingTable, RoutingRecord


class BalancerTest(unittest.TestCase):

    def test_weighted_random(self):
        services = ['aservice']
        route = RoutingRecord('aservice', hosts=['a', 'b'], weights=[2, 4])  # 0.333 / 0.666

        rtbl = RoutingTable()
        rtbl.list_services = unittest.mock.MagicMock(return_value=services)
        rtbl.get_routing = unittest.mock.MagicMock(return_value=route)

        balancer = WeightedRandomBalancer(rtbl)

        hosts = []
        for i in range(100):
            hosts.append(balancer.next_host('aservice'))

        cnt = Counter(hosts)

        self.assertAlmostEqual(33, cnt['a'], delta=15)
        self.assertAlmostEqual(66, cnt['b'], delta=15)

    def test_weighted_round_robin(self):
        services = ['aservice']
        route = RoutingRecord('aservice', hosts=['a', 'b', 'c'], weights=[1, 4, 5])  # 0.1 / 0.4 / 0.5

        rtbl = RoutingTable()
        rtbl.list_services = unittest.mock.MagicMock(return_value=services)
        rtbl.get_routing = unittest.mock.MagicMock(return_value=route)

        balancer = WeightedRoundRobinBalancer(rtbl)

        hosts = []
        for i in range(100):
            hosts.append(balancer.next_host('aservice'))

        cnt = Counter(hosts)

        self.assertAlmostEqual(10, cnt['a'], delta=1)
        self.assertAlmostEqual(40, cnt['b'], delta=1)
        self.assertAlmostEqual(50, cnt['c'], delta=1)

    def test_weighted_round_robin_multiple_services(self):
        services = ['aservice', 'bservice']

        def get_routing(service):
            if service == 'aservice':
                return RoutingRecord('aservice', hosts=['a', 'b', 'c'], weights=[1, 4, 5])
            elif service == 'bservice':
                return RoutingRecord('bservice', hosts=['a', 'd'], weights=[2, 8])
            else:
                raise ValueError

        rtbl = RoutingTable()
        rtbl.list_services = unittest.mock.MagicMock(return_value=services)
        rtbl.get_routing = unittest.mock.MagicMock(side_effect=get_routing)

        balancer = WeightedRoundRobinBalancer(rtbl)

        hosts = defaultdict(list)
        for _ in range(50):
            hosts['aservice'].append(balancer.next_host('aservice'))
        for _ in range(50):
            hosts['bservice'].append(balancer.next_host('bservice'))
        for _ in range(50):
            hosts['aservice'].append(balancer.next_host('aservice'))
            hosts['bservice'].append(balancer.next_host('bservice'))

        cnt_a = Counter(hosts['aservice'])
        self.assertAlmostEqual(10, cnt_a['a'], delta=1)
        self.assertAlmostEqual(40, cnt_a['b'], delta=1)
        self.assertAlmostEqual(50, cnt_a['c'], delta=1)

        cnt_b = Counter(hosts['bservice'])
        self.assertAlmostEqual(20, cnt_b['a'], delta=1)
        self.assertAlmostEqual(80, cnt_b['d'], delta=1)

    def test_weighted_round_robin_removing_hosts(self):
        rtbl = RoutingTable()
        rtbl.list_services = unittest.mock.MagicMock(return_value=['aservice'])
        rtbl.get_routing = unittest.mock.MagicMock(
            return_value=RoutingRecord('aservice', hosts=['a', 'b', 'c', 'd'], weights=[1, 1, 1, 1]))

        balancer = WeightedRoundRobinBalancer(rtbl)

        for _ in range(50):
            balancer.next_host('aservice')

        rtbl.get_routing = unittest.mock.MagicMock(
            return_value=RoutingRecord('aservice', hosts=['a', 'b', 'c'], weights=[1, 4, 5]))

        hosts = []
        for i in range(100):
            hosts.append(balancer.next_host('aservice'))

        cnt = Counter(hosts)

        self.assertAlmostEqual(10, cnt['a'], delta=1)
        self.assertAlmostEqual(40, cnt['b'], delta=1)
        self.assertAlmostEqual(50, cnt['c'], delta=1)

    def test_weighted_round_robin_adding_hosts(self):
        rtbl = RoutingTable()
        rtbl.list_services = unittest.mock.MagicMock(return_value=['aservice'])
        rtbl.get_routing = unittest.mock.MagicMock(
            return_value=RoutingRecord('aservice', hosts=['a', 'b'], weights=[1, 1]))

        balancer = WeightedRoundRobinBalancer(rtbl)

        for _ in range(50):
            balancer.next_host('aservice')

        rtbl.get_routing = unittest.mock.MagicMock(
            return_value=RoutingRecord('aservice', hosts=['a', 'b', 'c'], weights=[1, 4, 5]))

        hosts = []
        for i in range(100):
            hosts.append(balancer.next_host('aservice'))

        cnt = Counter(hosts)

        self.assertAlmostEqual(10, cnt['a'], delta=1)
        self.assertAlmostEqual(40, cnt['b'], delta=1)
        self.assertAlmostEqual(50, cnt['c'], delta=1)



import itertools
import json
import logging
import re
import sre_constants
from collections import Counter
from typing import List, Optional, Dict, Tuple

import pymq
from pymq.typing import deep_from_dict
from redis import Redis

from galileo.worker.api import RegisterWorkerCommand, ClientDescription, CreateClientCommand, ClientConfig, \
    StartTracingCommand, PauseTracingCommand, SetWorkloadCommand, StopWorkloadCommand

logger = logging.getLogger(__name__)


class ClusterController:

    def ping(self):
        raise NotImplementedError

    def discover(self):
        raise NotImplementedError

    def create_client(self, host: str, cfg: ClientConfig, num=1) -> List[ClientDescription]:
        raise NotImplementedError

    def create_clients(self, cfg: ClientConfig, num=1) -> List[ClientDescription]:
        raise NotImplementedError

    def register_worker(self, name: str, labels: Dict[str, str] = None):
        raise NotImplementedError

    def unregister_worker(self, name: str):
        raise NotImplementedError

    def list_workers(self, pattern: str = ''):
        raise NotImplementedError

    def list_workers_info(self, pattern: str = ''):
        raise NotImplementedError

    def register_client(self, client: ClientDescription):
        raise NotImplementedError

    def unregister_client(self, client_id: str):
        raise NotImplementedError

    def list_clients(self, worker: str = None) -> List[ClientDescription]:
        raise NotImplementedError

    def get_client_description(self, client_id: str) -> Optional[ClientDescription]:
        raise NotImplementedError

    def start_tracing(self):
        raise NotImplementedError

    def stop_tracing(self):
        raise NotImplementedError

    def set_workload(self, client_id, ia=None, n: int = None):
        raise NotImplementedError

    def stop_workload(self, client_id):
        raise NotImplementedError


def pack(n, workers, nr_clients):
    """
    Takes an input a list of workers, and a list of the current number of clients on each worker (same order). Returns a
    dictionary {worker: nr}, where nr is the number of clients that should be spawned on the worker.

    :param n: the number of new clients to distribute
    :param workers: the list of workers
    :param nr_clients: the number of current clients on each worker
    :return: a dictionary {worker: nr}
    """

    def packer():
        cur = list(nr_clients)

        while True:
            index_min = min(range(len(cur)), key=cur.__getitem__)
            cur[index_min] += 1
            yield workers[index_min]

    target_workers = list(itertools.islice(packer(), n))
    return dict(Counter(target_workers))


class RedisClusterController(ClusterController):
    worker_key = 'galileo:workers'
    worker_clients_key = 'galileo:worker:%s:clients'
    client_key = 'galileo:client:%s'

    def __init__(self, rds, eventbus=None) -> None:
        super().__init__()
        self.rds = rds
        self.eventbus = eventbus or pymq

    def ping(self):
        stub = self.eventbus.stub('WorkerDaemon.ping', multi=True, timeout=2)
        return stub()

    def discover(self):
        for worker in self.list_workers():
            self.rds.delete(self.worker_clients_key % worker)

        self.rds.delete(self.worker_key)
        return self.eventbus.publish(RegisterWorkerCommand())

    def register_worker(self, name: str, labels: Dict[str, str] = None):
        logger.info('registering worker %s', name)
        self.rds.sadd(self.worker_key, name)
        if labels is not None and len(labels) > 0:
            self.rds.hmset(name, labels)

    def unregister_worker(self, name: str):
        logger.info('unregistering worker %s', name)
        self.rds.srem(self.worker_key, name)
        self.rds.delete(self.worker_clients_key % name)

    def list_workers(self, pattern: str = ''):
        workers = self.rds.smembers(self.worker_key)

        if not pattern:
            return workers

        try:
            return [worker for worker in workers if re.search('^%s$' % pattern, worker)]
        except sre_constants.error as e:
            raise ValueError('Invalid pattern %s: %s' % (pattern, e))

    def list_workers_info(self, pattern: str = '') -> List[Tuple[str, Dict[str, str]]]:
        workers = self.list_workers(pattern)
        items = []
        for worker in workers:
            data = self.rds.hgetall(worker)
            items.append((worker, data))
        return items

    def create_client(self, host: str, cfg: ClientConfig, num=1) -> List[ClientDescription]:
        cmd = CreateClientCommand(host, cfg, num)
        stub = self.eventbus.stub(f'WorkerDaemon.create_client:{host}', timeout=3)

        result = stub(cmd)
        # TODO: error handling
        return [deep_from_dict(d, ClientDescription) for d in result]

    def create_clients(self, cfg: ClientConfig, num=1) -> List[ClientDescription]:
        worker_client_count = self.count_worker_clients(cfg.worker_labels)
        workers = list(worker_client_count.keys())

        nr_clients = list(worker_client_count.values())

        schedule = pack(num, workers, nr_clients)

        clients = list()
        for worker, nr in schedule.items():
            created = self.create_client(worker, cfg, nr)
            clients.extend(created)

        return clients

    def register_client(self, client: ClientDescription):
        # add the client_id to the worker's clients
        worker_clients_key = self.worker_clients_key % client.worker
        self.rds.sadd(worker_clients_key, client.client_id)

        # set the client description
        client_key = self.client_key % client.client_id
        self.rds.set(client_key, serialize_client_description(client))

    def unregister_client(self, client_id: str):
        client = self.get_client_description(client_id)

        if not client:
            return

        # remove the client description
        client_key = self.client_key % client_id
        self.rds.delete(client_key)

        # remove client_id from the worker's clients
        worker_clients_key = self.worker_clients_key % client.worker
        self.rds.srem(worker_clients_key, client.client_id)

    def list_clients(self, worker: str = None) -> List[ClientDescription]:
        # TODO: error handling ¯\_(ツ)_/¯

        rds: Redis = self.rds

        if worker is not None:
            client_ids = rds.smembers(self.worker_clients_key % worker)
        else:
            # first get the keys that holds the client ids hosted on the worker
            keys = set(rds.scan_iter(self.worker_clients_key % '*'))
            # the union of those sets contains all client_ids we're looking for
            if not keys:
                return []

            client_ids = rds.sunion(keys)

        if not client_ids:
            return []

        client_keys = [self.client_key % client_id for client_id in client_ids]
        docs = rds.mget(client_keys)

        descriptions = [deserialize_client_description(doc) for doc in docs if doc is not None]

        return descriptions

    def get_client_description(self, client_id: str) -> Optional[ClientDescription]:
        key = self.client_key % client_id
        doc = self.rds.get(key)

        if doc is None:
            return None

        return deserialize_client_description(doc)

    def start_tracing(self):
        return self.eventbus.publish(StartTracingCommand())

    def stop_tracing(self):
        return self.eventbus.publish(PauseTracingCommand())

    def set_workload(self, client_id, ia=None, n: int = None):
        if ia is None and n is None:
            raise ValueError('need interarrival or number of messages')

        dist, params = 'constant', None

        if isinstance(ia, (int, float)):
            dist, params = 'constant', (ia,)
        elif isinstance(ia, tuple):
            dist, params = ia[0], ia[1:]

        cmd = SetWorkloadCommand(client_id, num=n, distribution=dist, parameters=params)
        return self.eventbus.publish(cmd)

    def stop_workload(self, client_id):
        return self.eventbus.publish(StopWorkloadCommand(client_id))

    def count_worker_clients(self, worker_labels: Dict[str, str] = None):
        workers_labels = self.list_workers_info()
        workers = list(map(lambda x: x[0], workers_labels))
        if worker_labels is not None:
            for worker, labels in workers_labels:
                for k, v in worker_labels.items():
                    if labels[k] != v:
                        workers.remove(worker)
                        break

        if len(workers) == 0:
            raise IndexError('No workers found that match labels.')

        pipe = self.rds.pipeline()
        for worker in workers:
            pipe.scard(self.worker_clients_key % worker)

        nr_clients = pipe.execute()
        return dict(zip(workers, nr_clients))


def serialize_client_description(obj: ClientDescription):
    return json.dumps(obj)


def deserialize_client_description(doc: str):
    return deep_from_dict(json.loads(doc), cls=ClientDescription)



import unittest

from galileo.apps.app import DefaultAppClient
from galileo.worker.client import AppClientRequestFactory


class DefaultAppClientTest(unittest.TestCase):

    def test_with_factory(self):
        client = DefaultAppClient()
        factory = AppClientRequestFactory('myservice', client)

        req = factory.create_request()

        self.assertEqual('myservice', req.service)
        self.assertEqual('/', req.path)
        self.assertEqual('get', req.method)
        self.assertEqual({}, req.kwargs)

    def test_with_parameters_and_factory(self):
        client = DefaultAppClient({'method': 'post', 'path': '/foo', 'kwargs': {'data': '500'}})
        factory = AppClientRequestFactory('myservice', client)

        req = factory.create_request()

        self.assertEqual('myservice', req.service)
        self.assertEqual('/foo', req.path)
        self.assertEqual('post', req.method)
        self.assertEqual({'data': '500'}, req.kwargs)



from typing import List

from galileodb.model import Experiment, Telemetry, RequestTrace

from tests.api import ResourceTest


class TestExperimentResource(ResourceTest):

    def test_get(self):
        exp = Experiment('id1', 'name1', 'creator1', 123, 456, 789, 'FINISH')
        self.db_resource.db.save_experiment(exp)
        result = self.simulate_get(f'/api/experiments/id1')

        self.assertIsNotNone(result.json)
        self.assertEqual(result.json, exp.__dict__)

    def test_delete(self):
        exp_id = 'id1'
        exp = Experiment(exp_id, 'name1', 'creator1', 1, 20, 7, 'FINISH')
        telemetry: List[Telemetry] = [Telemetry(1, 'metric1', 'node1', 1, exp_id),
                                      Telemetry(2, 'metric2', 'node2', 2, exp_id)]
        traces: List[RequestTrace] = [RequestTrace('req1', 'client1', 'service1', 2, 2, 3),
                                      RequestTrace('req2', 'client2', 'service2', 6, 5, 4)]
        db = self.db_resource.db
        db.save_experiment(exp)
        db.save_telemetry(telemetry)
        db.save_traces(traces)
        db.touch_traces(exp)

        self.simulate_delete('/api/experiments/id1')

        self.assertEqual(len(db.find_all()), 0)

        traces_fetched = db.db.fetchall('SELECT * FROM traces')
        telemetry_fetched = db.db.fetchall('SELECT * FROM telemetry')
        self.assertEqual(len(traces_fetched), 0)
        self.assertEqual(len(telemetry_fetched), 0)
        pass



import os
import threading
import unittest
from http.server import HTTPServer, BaseHTTPRequestHandler
from queue import Queue
from typing import List

import pymq
from galileodb.model import RequestTrace
from galileodb.recorder.traces import RedisTraceRecorder
from galileodb.trace import TraceWriter
from pymq.provider.redis import RedisConfig
from timeout_decorator import timeout_decorator

from galileo.controller import RedisClusterController
from galileo.routing import RedisRoutingTable, RoutingRecord
from galileo.shell.shell import Galileo
from galileo.util import poll
from galileo.worker.daemon import WorkerDaemon
from tests.testutils import assert_poll, RedisResource


class TestIntegration(unittest.TestCase):
    redis_resource = RedisResource()

    def setUp(self) -> None:
        # Create redis server
        self.redis_resource.setUp()
        self.rds = self.redis_resource.rds
        self.rds.flushall()

        self.eventbus = pymq.init(RedisConfig(self.rds))
        self.ctrl = RedisClusterController(self.rds)
        self.rtbl = RedisRoutingTable(self.rds)

    def tearDown(self) -> None:
        pymq.shutdown()
        self.redis_resource.tearDown()

    @timeout_decorator.timeout(30)
    def test_single_client_scenario(self):
        os.environ['galileo_redis_host'] = 'file://' + self.redis_resource.tmpfile
        os.environ['galileo_trace_logging'] = 'redis'
        os.environ['galileo_router_type'] = 'SymmetryHostRouter'

        # setup up http app
        class HttpHandler(BaseHTTPRequestHandler):
            def do_GET(self):
                self.send_response(200)
                self.send_header('Content-type', 'text/html')
                self.end_headers()
                self.wfile.write(b'ok')
                return

        httpd = HTTPServer(('localhost', 31523), HttpHandler)
        t_http = threading.Thread(target=httpd.serve_forever, daemon=True)
        t_http.start()

        # start a worker instance
        worker = WorkerDaemon()
        t_worker = threading.Thread(target=worker.run, daemon=True)
        t_worker.start()

        # start traces subscriber
        traces = Queue()

        class QueueTraceWriter(TraceWriter):
            def write(self, data: List[RequestTrace]):
                for d in data:
                    traces.put(d)

        recorder = RedisTraceRecorder(self.rds, exp_id='exp_01', writer=QueueTraceWriter(), flush_every=1)
        recorder.start()

        try:
            # set rtbl record
            self.rtbl.set_routing(RoutingRecord('myservice', ['localhost:31523'], [1]))

            g = Galileo(self.ctrl)

            # test a single request and make sure everything works so far
            response = g.request('myservice')
            self.assertEqual('ok', response.text)

            # wait for worker to be up
            self.assertEqual(1, len(g.ping()))

            # start tracing
            g.start_tracing()

            # spawn a client group
            c = g.spawn('myservice')

            # wait for the client to appear
            assert_poll(lambda: pymq.stub('Client.get_info')() is not None, msg='client did not appear')

            # make sure client was registered
            assert_poll(lambda: len(self.ctrl.list_clients()) >= 1, msg='client was not registered')

            # send three requests and wait for them to finish
            c.request(n=3).wait(2)

            # close clients
            c.close()

            # wait for them to disappear
            assert_poll(lambda: len(self.ctrl.list_clients()) == 0, msg='client was not closed')

            # should flush traces to redis
            g.stop_tracing()

            # check the first trace fully
            t: RequestTrace = traces.get(timeout=5)
            self.assertEqual(200, t.status)
            self.assertEqual('ok', t.response)
            headers = '"content-type": ["Content-type", "text/html"]'
            self.assertTrue(headers in t.headers)
            self.assertEqual('myservice', t.service)
            self.assertEqual('localhost:31523', t.server)

            # check that the others are successful
            t: RequestTrace = traces.get(timeout=2)
            self.assertEqual(200, t.status)
            t: RequestTrace = traces.get(timeout=2)
            self.assertEqual(200, t.status)

            self.assertEqual(0, traces.qsize(), 'expected trace queue to have exactly three entries')

        finally:
            worker.close()
            httpd.shutdown()
            httpd.server_close()
            recorder.stop(2)

        recorder.join(2)
        t_worker.join(2)
        t_http.join(2)

        # wait for client to be gone to safely shutdown redis
        poll(lambda: pymq.stub('Client.get_info', timeout=0.5)() is None, timeout=2)



import logging
import time

import falcon
import redis
from galileodb import ExperimentDatabase
from galileodb.model import WorkloadConfiguration, ExperimentConfiguration, Experiment, generate_experiment_id

from galileo.apps.repository import Repository
from galileo.webapp.repository import RepositoryResource
from galileo.controller import ExperimentController, CancelError, ClusterController
from galileo.experiment.service import ExperimentService
from galileo.util import to_seconds

logger = logging.getLogger(__name__)


class AppContext:
    rds: redis.Redis
    cctrl: ClusterController
    ectrl: ExperimentController
    exp_db: ExperimentDatabase
    exp_service: ExperimentService
    repository: Repository


def setup(api: falcon.API, context: AppContext):
    api.add_route('/api', ApiResource(api))
    api.add_route('/api/worker', WorkerResource(context.cctrl))

    api.add_route('/api/services', ServicesResource())
    api.add_route('/api/experiments', ExperimentsResource(context.cctrl, context.ectrl, context.exp_service))
    api.add_route('/api/experiments/{exp_id}', ExperimentResource(context.exp_service, context.ectrl))

    repo = RepositoryResource(context.repository)
    api.add_route('/api/apps', repo)
    api.add_route('/api/apps/{app_name}', repo, suffix='info')
    api.add_route('/api/apps/{app_name}/download', repo, suffix='download')


class ServicesResource:
    def on_get(self, req, resp):
        # TODO: rethink this
        services = [
            {
                'name': 'mxnet-model-server'
            }
        ]

        resp.media = services


class WorkerResource:

    def __init__(self, ctrl: ClusterController):
        self.ctrl = ctrl

    def on_get(self, req, resp):
        resp.media = list(self.ctrl.list_workers())


class ExperimentsResource:

    def __init__(self, cctrl: ClusterController, ectrl: ExperimentController, exp_service: ExperimentService):
        self.cctrl = cctrl
        self.ectrl = ectrl
        self.exp_service = exp_service

    def on_get(self, req: falcon.Request, resp):
        logger.debug('fetching all experiments')
        experiments = self.exp_service.find_all()
        logger.debug(f"found {len(experiments)} experiments")
        resp.media = [exp.__dict__ for exp in experiments]

    """
    here's an example request:

    {
        'experiment': { # experiment is optional, all attributes can be generated
            'name': 'my_experiment',
            'creator': 'research star'
        },
        'configuration': {
            'duration': '10s',
            'interval': '2s',
            'workloads': [
                {
                    'service': 'alexnet',
                    'ticks': [1, 2, 3, 1, 1],  # len must be duration / interval
                    'clients_per_host': 2  # optional, will be set to 3 by default,
                    'client_parameters': {"size": "small", "type": "cat"} # optional
                },
                # ...
            ]
        }
    }
    """

    def on_post(self, req: falcon.Request, resp):
        if not self.cctrl.list_workers():
            raise falcon.HTTPServiceUnavailable('no available hosts to execute the experiment')

        doc = req.media

        exp = doc['experiment'] if 'experiment' in doc else dict()
        if 'id' not in exp:
            exp['id'] = generate_experiment_id()

        exp = Experiment(**exp)
        exp.created = time.time()
        exp.status = 'QUEUED'
        self.exp_service.save(exp)

        logger.debug('deserialized experiment %s', exp)

        workloads = [WorkloadConfiguration(**workload) for workload in doc['configuration']['workloads']]
        duration = to_seconds(doc['configuration']['duration'])
        interval = to_seconds(doc['configuration']['interval'])
        config = ExperimentConfiguration(duration, interval, workloads)
        logger.debug('deserialized experiment config %s', config)
        logger.debug('queuing experiment with id %s', exp.id)
        self.ectrl.queue(config, exp)

        resp.media = exp.id


class ExperimentResource:

    def __init__(self, exp_service: ExperimentService, ectrl: ExperimentController):
        self.exp_service = exp_service
        self.ectrl = ectrl

    def on_get(self, req, resp, exp_id):
        logger.debug('finding experiment %s', exp_id)
        experiment = self.exp_service.find(exp_id)

        if not experiment:
            raise falcon.HTTPNotFound()

        resp.media = experiment.__dict__

    def on_delete(self, req, resp, exp_id):
        logger.debug('deleting experiment %s', exp_id)

        try:
            exp: Experiment = self.exp_service.find(exp_id)
            if exp is None:
                raise falcon.HTTPNotFound()

            if exp.status.lower() == 'queued':
                cancelled = False
                try:
                    cancelled = self.ectrl.cancel(exp_id)
                    self.exp_service.delete(exp_id)
                except CancelError:
                    try:
                        cancelled = self.ectrl.cancel(exp_id)
                        self.exp_service.delete(exp_id)
                    except CancelError:
                        logger.error(f'Cancellation of exp with id {exp_id} failed two times')

                if cancelled:
                    logger.info(f"Experiment {exp_id} was removed from queue")
                else:
                    logger.info(f"Experiment {exp_id} was not removed from queue")
            else:
                self.exp_service.delete(exp_id)
        except ValueError:
            raise falcon.HTTPNotFound()

        resp.media = exp_id


class CORSComponent(object):
    """
    CORS preprocessor from the Falcon documentation.
    """

    def process_response(self, req, resp, resource, req_succeeded):
        resp.set_header('Access-Control-Allow-Origin', '*')

        if (req_succeeded
                and req.method == 'OPTIONS'
                and req.get_header('Access-Control-Request-Method')
        ):
            # NOTE(kgriffs): This is a CORS preflight request. Patch the
            #   response accordingly.

            allow = resp.get_header('Allow')
            resp.delete_header('Allow')

            allow_headers = req.get_header(
                'Access-Control-Request-Headers',
                default='*'
            )

            resp.set_headers((
                ('Access-Control-Allow-Methods', allow),
                ('Access-Control-Allow-Headers', allow_headers),
                ('Access-Control-Max-Age', '86400'),  # 24 hours
            ))


class ApiResource:
    api: falcon.API

    def __init__(self, api, prefix='/api') -> None:
        super().__init__()
        self.api = api
        self.prefix = prefix

    def on_get(self, req: falcon.Request, resp):
        resp.media = self.get_all_routes(req, self.api)

    def get_all_routes(self, req, api):
        routes_list = []

        def get_children(node):
            if len(node.children):
                for child_node in node.children:
                    get_children(child_node)
            else:
                if node.uri_template.startswith(self.prefix):
                    routes_list.append({
                        'uri': node.uri_template,
                        'type': str(type(node.resource).__name__)
                    })

        [get_children(node) for node in api._router._roots]

        routes_list.sort(key=lambda k: k['uri'])

        return routes_list



import unittest

from tests.api import ResourceTest


class TestServiceResource(ResourceTest):
    def test_get_services(self):
        services = [
            {
                'name': 'mxnet-model-server'
            }
        ]

        result = self.simulate_get('/api/services')

        self.assertEqual(result.json, services)


if __name__ == '__main__':
    unittest.main()



import json
import logging
import signal
import threading
import time
from concurrent.futures import ThreadPoolExecutor
from multiprocessing import Queue
from queue import Full
import pymq
import requests
from galileodb.model import RequestTrace
from pymq.provider.redis import RedisEventBus

from galileo import util
from galileo.apps.app import AppClient, DefaultAppClient
from galileo.routing import ServiceRequest
from galileo.worker.api import ClientDescription, ClientConfig, ClientInfo, SetWorkloadCommand, StopWorkloadCommand, \
    WorkloadDoneEvent
from galileo.worker.context import Context
from galileo.worker.random import create_sampler

logger = logging.getLogger(__name__)


def constant(mean):
    while True:
        yield mean


def limiter(limit, gen):
    for _ in range(limit):
        yield next(gen)


def create_interarrival_generator(cmd: SetWorkloadCommand, ctx: Context):
    if not cmd.distribution or cmd.distribution == 'constant':
        if cmd.parameters:
            gen = constant(*cmd.parameters)
        else:
            gen = constant(0)
    else:
        logger.debug(f'generating sampler {cmd.distribution} with client_id {cmd.client_id}')
        gen = create_sampler(cmd.distribution, cmd.parameters, ctx, client_id=cmd.client_id)

    if cmd.num:
        return limiter(cmd.num, gen)
    else:
        return gen


class RequestGenerator:
    DONE = object()

    def __init__(self, factory, ctx=None) -> None:
        super().__init__()
        self.factory = factory
        self.ctx = ctx

        self._closed = False

        self._gen = None
        self._gen_lock = threading.Condition()

    def close(self):
        with self._gen_lock:
            logger.debug('closing generator %s', self)
            self._closed = True
            self._gen_lock.notify_all()

    def set_workload(self, cmd: SetWorkloadCommand):
        with self._gen_lock:
            gen = create_interarrival_generator(cmd, self.ctx)
            self._gen = gen
            self._gen_lock.notify_all()

    def pause(self):
        with self._gen_lock:
            self._gen = None

    def _next_interarrival(self):
        gen = self._gen
        if gen is not None:
            return next(gen)

        with self._gen_lock:
            if self._gen is None:  # set_rps may already have been called and notified has_gen
                logger.debug('generator paused %s', self)
                self._gen_lock.wait()
                if self._closed or not self._gen:
                    raise InterruptedError

            logger.debug('generator resumed %s', self)

            return next(self._gen)

    def run(self):
        logger.debug('running request generator %s', self)

        factory = self.factory
        last_log_time = time.time()

        while not self._closed:
            try:
                a = self._next_interarrival()  # may block until a generator is available
                if a > 0:
                    time.sleep(a)
                else:
                    raise ValueError(f'ia time was {a}, which is invalid and leads to massive crashes!')
                if time.time() - last_log_time >= 5 and a != 0:
                    logger.debug(f'Last ia time indicated target request rate of {str(1/a)}rps')
                    last_log_time = time.time()
            except StopIteration:
                with self._gen_lock:
                    self._gen = None
                yield RequestGenerator.DONE
                continue
            except InterruptedError:
                break

            yield factory()


class AppClientRequestFactory:

    def __init__(self, service: str, client: AppClient) -> None:
        super().__init__()
        self.service = service
        self.client = client

    def create_request(self) -> ServiceRequest:
        req = self.client.next_request()
        service_request = ServiceRequest(self.service, req.endpoint, req.method, **req.kwargs)

        if logger.isEnabledFor(logging.DEBUG):
            logger.debug('client %s created request %s', self.client.name, service_request.__dict__)

        return service_request

    def __call__(self, *args, **kwargs):
        return self.create_request()


class Client:

    def __init__(self, ctx: Context, trace_queue: Queue, description: ClientDescription, eventbus=None) -> None:
        super().__init__()
        self.ctx = ctx
        self.description = description
        self.client_id = description.client_id
        self.cfg = description.config
        self.traces = trace_queue
        self.eventbus = eventbus or pymq

        self.router = ctx.create_router()
        self.request_generator = RequestGenerator(self._create_request_factory(), self.ctx)

        # used for generating request ids
        self.client_uuid = util.uuid()[-10:]
        self.request_counter = 0

        # used for statistics
        self.failed_counter = 0

        # expose methods
        self.eventbus.subscribe(self._on_set_workload_command)
        self.eventbus.subscribe(self._on_stop_workload_command)
        self.eventbus.expose(self.get_info, 'Client.get_info')
        self.request_executor = ThreadPoolExecutor(max_workers=50)

    def _create_request_id(self, _: ServiceRequest) -> str:
        self.request_counter += 1
        return self.client_uuid + ":" + str(self.request_counter)

    def get_info(self) -> ClientInfo:
        return ClientInfo(self.description, self.request_counter, self.failed_counter)

    def perform_request(self, request):
        client_id = self.client_id
        traces = self.traces
        router = self.router
        if request is RequestGenerator.DONE:
            self.eventbus.publish(WorkloadDoneEvent(self.client_id))
            return

        logger.debug('client %s processing request %s', client_id, request)
        request.client_id = client_id
        request.request_id = self._create_request_id(request)

        try:
            request.sent = -1  # will be updated by router
            response: requests.Response = router.request(request)
            host = response.url.split("//")[-1].split("/")[0].split('?')[0]

            t = RequestTrace(
                request_id=request.request_id,
                client=client_id,
                service=request.service,
                created=request.created,
                sent=request.sent,
                done=time.time(),
                status=response.status_code,
                server=host,
                response=response.text.strip(),
                headers=json.dumps(dict(response.headers))
            )
        except Exception as e:
            if logger.isEnabledFor(logging.DEBUG):
                logger.exception('error while handling request %s', request)
            else:
                logger.error('error while handling request %s: %s', request, e)

            t = RequestTrace(
                request_id=request.request_id,
                client=client_id,
                service=request.service,
                created=request.created,
                sent=request.sent,
                done=time.time(),
                status=-1
            )

        if t.status < 0 or t.status >= 300:
            self.failed_counter += 1

        try:
            traces.put_nowait(t)
        except Full:
            pass

    def run(self):
        client_id = self.client_id

        rgen = self.request_generator.run()
        last_log_update = time.time()
        dispatches_since_last_log_update = 0
        try:
            while True:
                logger.debug("client %s waiting for next request", client_id)
                try:
                    request = next(rgen)
                    self.request_executor.submit(self.perform_request, request)
                    dispatches_since_last_log_update += 1
                    if time.time() - last_log_update >= 1:
                        logger.debug(f'queued {dispatches_since_last_log_update} requests last second')
                        dispatches_since_last_log_update = 0
                        last_log_update = time.time()
                except StopIteration:
                    break

        except KeyboardInterrupt:
            return
        except:
            logger.exception("error during read loop in client %s", client_id)
        finally:
            self.request_executor.shutdown(wait=False)

    def close(self):
        self.request_generator.close()
        self.eventbus.unsubscribe(self._on_set_workload_command)
        self.eventbus.unsubscribe(self._on_stop_workload_command)
        self.eventbus.unexpose('Client.get_info')

    def _on_set_workload_command(self, cmd: SetWorkloadCommand):
        if cmd.client_id != self.client_id:
            return

        self.request_generator.set_workload(cmd)

    def _on_stop_workload_command(self, cmd: StopWorkloadCommand):
        if cmd.client_id != self.client_id:
            return

        self.request_generator.pause()

    def _create_request_factory(self):
        if self.cfg.client:
            app_loader = self.ctx.create_app_loader()
            app = app_loader.load(self.cfg.client, self.cfg.parameters)
        else:
            app = DefaultAppClient(self.cfg.parameters)

        return AppClientRequestFactory(self.cfg.service, app)

    def __str__(self):
        return 'Client{client_id=%s}' % self.client_id


def single_request(cfg: ClientConfig, ctx=None, router_type=None) -> requests.Response:
    ctx = ctx or Context()

    if cfg.client:
        app_loader = ctx.create_app_loader()
        app = app_loader.load(cfg.client, cfg.parameters)
    else:
        app = DefaultAppClient(cfg.parameters)

    factory = AppClientRequestFactory(cfg.service, app)

    router = ctx.create_router(router_type)
    return router.request(factory.create_request())


def run(ctx: Context, trace_queue: Queue, description: ClientDescription):
    logger.info('starting new client process %s', description)

    bus = RedisEventBus(rds=ctx.create_redis())
    bus_thread = threading.Thread(target=bus.run)
    bus_thread.start()

    client = Client(ctx, trace_queue, description, eventbus=bus)

    def handler(signum, frame):
        logger.debug('client %s received signal %s', client.client_id, signum)
        client.close()
        raise KeyboardInterrupt

    signal.signal(signal.SIGINT, handler)
    signal.signal(signal.SIGTERM, handler)

    try:
        logger.info("%s starting", client)
        client.run()
    except KeyboardInterrupt:
        pass

    logger.debug('shutting down eventbus')
    bus.close()
    bus_thread.join(2)

    logger.info("%s exitting", client)



import random
import time
import logging
from galileo.worker.context import Context

logger = logging.getLogger(__name__)

class InvalidDistributionException(Exception):
    pass


def constant(x):
    return x


distributions = {
    'constant': constant,
    'uniform': random.uniform,
    'triangular': random.triangular,
    'normalvariate': random.normalvariate,
    'lognormvariate': random.lognormvariate,
    'expovariate': random.expovariate,
    'vonmisesvariate': random.vonmisesvariate,
    'gammavariate': random.gammavariate,
    'gauss': random.gauss,
    'betavariate': random.betavariate,
    'paretovariate': random.paretovariate,
    'weibullvariate': random.weibullvariate
}


def pre_recorded_profile(ctx: Context, list_key: str):
    rds = ctx.create_redis()
    start = time.time()
    values = rds.lrange(list_key, 0, rds.llen(list_key))
    rds.delete(list_key)
    # Values need to be reversed since we originally treated this like a stack
    values.reverse()
    end = time.time()
    logger.debug(f'loaded {len(values)} ia values in {end - start}s from redis')
    for value in values:
        yield float(value)
    logger.info('done')

def create_sampler(distribution: str, args: tuple, ctx: Context, client_id=None):
    """
    Creates a generator for the given distribution with the given arguments.

    :param distribution: the distribution, e.g., 'lognormvariate'
    :param args: the arguments, e.g., (0.5, 1)
    :param ctx: context object
    :return: a generator
    """
    print("create sampler")
    if distribution == 'prerecorded':
        yield from pre_recorded_profile(ctx, client_id)
    else:
        if distribution not in distributions:
            raise InvalidDistributionException('unknown distribution ' + distribution)

        try:
            fn = distributions[distribution]
        except TypeError as e:
            raise InvalidDistributionException('invalid distribution parameters: ' + str(e))

        if args is None:
            args = []

        while True:
            yield fn(*args)



import math
import time

from galileodb.model import ExperimentConfiguration

from galileo.controller import ClusterController
from galileo.shell.shell import Galileo


def run_experiment(ctrl: ClusterController, exp: ExperimentConfiguration):
    g = Galileo(ctrl)
    workload_clients = {}

    for workload in exp.workloads:
        workload_clients[workload] = g.spawn(workload.service, num=workload.clients_per_host, client=workload.client,
                                             parameters=workload.client_parameters)

    ticks = int(math.ceil(exp.duration / exp.interval))

    for t in range(ticks):
        for workload, clients in workload_clients.items():
            service_rps = workload.ticks[t]
            clients.rps(service_rps / workload.clients_per_host)

        time.sleep(exp.interval)

    for clients in workload_clients.values():
        clients.rps(0)

    for clients in workload_clients.values():
        clients.close()



from time import sleep

from tests.api import ResourceTest


class TestHostResource(ResourceTest):

    def test_get_empty_hosts(self):
        result = self.simulate_get('/api/worker')

        self.assertEqual(result.json, [])

    def test_get_worker(self):
        hosts = ['host1', 'host2']

        self.ctx.cctrl.register_worker('host1')
        self.ctx.cctrl.register_worker('host2')

        # sleep to publish events and add hosts to redis
        sleep(0.5)

        result = self.simulate_get('/api/worker')
        self.assertIsNotNone(result.json, 'response must not be empty')
        self.assertCountEqual(result.json, hosts)



import unittest

from galileo.util import to_seconds


class TestUtil(unittest.TestCase):
    def test_to_seconds(self):
        self.assertEqual(0, to_seconds(''))
        self.assertEqual(1, to_seconds('1'))
        self.assertEqual(10, to_seconds('10s'))
        self.assertEqual(60, to_seconds('1m'))
        self.assertEqual(600, to_seconds('10m'))
        self.assertEqual(610, to_seconds('10m 10s'))



import argparse
import logging
import os
import signal

import pymq
from pymq.provider.redis import RedisConfig

from galileo.worker.context import Context
from galileo.worker.daemon import WorkerDaemon

logger = logging.getLogger(__name__)


def main():
    signal.signal(signal.SIGINT, signal_handler)
    signal.signal(signal.SIGTERM, signal_handler)

    parser = argparse.ArgumentParser()
    parser.add_argument('--logging', required=False,
                        help='set log level (DEBUG|INFO|WARN|...) to activate logging',
                        default=os.getenv('galileo_logging_level'))
    args = parser.parse_args()

    if args.logging:
        logging.basicConfig(level=logging._nameToLevel[args.logging])

    context = Context()

    redis = require_redis(context)
    eventbus = pymq.init(RedisConfig(redis))

    # run worker
    worker = WorkerDaemon(context, eventbus=eventbus)

    try:
        logger.info('starting experiment worker %s', worker.name)
        worker.run()
    except KeyboardInterrupt:
        pass
    finally:
        worker.close()

    logger.info('done, bye')


def require_redis(context):
    try:
        redis = context.create_redis()
        redis.randomkey()  # guard for redis connection
        return redis
    except Exception as e:
        print('Could not initiate Redis connection:', e)

    exit(1)


def signal_handler(signum, frame):
    raise KeyboardInterrupt


if __name__ == '__main__':
    main()



import atexit
import logging
import os
import time
from socket import gethostname
from typing import MutableMapping, List, Dict

import redis
import requests
from galileodb import ExperimentDatabase
from galileodb.factory import create_experiment_database_from_env
from galileodb.trace import TraceLogger, TraceWriter, FileTraceWriter, RedisTopicTraceWriter, DatabaseTraceWriter

from galileo.apps.loader import AppClientLoader, AppClientDirectoryLoader, AppRepositoryFallbackLoader
from galileo.apps.repository import RepositoryClient
from galileo.routing import Router, ServiceRequest, ServiceRouter, HostRouter, StaticRouter, RedisRoutingTable, \
    ReadOnlyListeningRedisRoutingTable, WeightedRoundRobinBalancer

logger = logging.getLogger(__name__)


class Context:
    """
    Factory for various worker services. Below are the environment variables that can be set:

    - Logging
        - galileo_log_level (DEBUG|INFO|WARN| ... )

    - Redis connection:
        - galileo_redis_host (localhost)
        - galileo_redis_port (6379)
        - galileo_redis_password (None)

    - Trace logging:
        - galileo_trace_logging: file|redis|sql
        - mysql:
            - galileo_expdb_driver: sqlite|mysql
            - sqlite:
                - galileo_expdb_sqlite_path ('/tmp/galileo_sqlite')
            - mysql:
                - galileo_expdb_mysql_host (localhost)
                - galileo_expdb_mysql_port (3307)
                - galileo_expdb_mysql_user
                - galileo_expdb_mysql_password
                - galileo_expdb_mysql_db

    - Request router
        - galileo_router_type: SymmetryServiceRouter|SymmetryHostRouter|StaticRouter|DebugRouter
            - StaticRouter:
                - galileo_router_static_host (http://localhost)

    - Client app loader:
        - galileo_apps_dir ('./apps')
        - galileo_apps_repository ('http://localhost:5001')
    """

    def __init__(self, env: MutableMapping = os.environ) -> None:
        super().__init__()
        self.env = env

    def getenv(self, *args, **kwargs):
        return self.env.get(*args, **kwargs)

    def keys(self, prefix: str = 'galileo_') -> List[str]:
        if prefix is None:
            return list(self.env.keys())
        else:
            return list(filter(lambda x: x.startswith('galileo_'), self.env.keys()))

    def items(self, prefix: str = 'galileo_') -> Dict[str, str]:
        if prefix is None:
            return dict(self.env.values())
        else:
            values = {}
            keys = self.keys(prefix)
            for key in keys:
                value = self.getenv(key)
                values[key] = value
        return values

    @property
    def worker_name(self):
        return self.env.get('galileo_worker_name', gethostname())

    def create_trace_writer(self) -> TraceWriter:
        trace_logging = self.env.get('galileo_trace_logging')
        logger.debug('trace logging: %s', trace_logging or 'None')

        if not trace_logging:
            return None
        elif trace_logging == 'file':
            return FileTraceWriter(self.worker_name)
        elif trace_logging == 'redis':
            return RedisTopicTraceWriter(self.create_redis())
        elif trace_logging == 'sql':
            return DatabaseTraceWriter(self.create_exp_db())
        else:
            raise ValueError('Unknown trace logging type %s' % trace_logging)

    def create_trace_logger(self, trace_queue, start=True) -> TraceLogger:
        writer = self.create_trace_writer()
        return TraceLogger(trace_queue, writer, start)

    def create_router(self, router_type=None):
        if router_type is None:
            router_type = self.env.get('galileo_router_type', 'CachingSymmetryHostRouter')

        if router_type == 'SymmetryServiceRouter':
            rtable = RedisRoutingTable(self.create_redis())
            balancer = WeightedRoundRobinBalancer(rtable)
            return ServiceRouter(balancer)
        elif router_type == 'CachingSymmetryServiceRouter':
            rtable = ReadOnlyListeningRedisRoutingTable(self.create_redis())
            rtable.start()
            atexit.register(rtable.stop, timeout=2)
            balancer = WeightedRoundRobinBalancer(rtable)
            return ServiceRouter(balancer)
        elif router_type == 'SymmetryHostRouter':
            rtable = RedisRoutingTable(self.create_redis())
            balancer = WeightedRoundRobinBalancer(rtable)
            return HostRouter(balancer)
        elif router_type == 'CachingSymmetryHostRouter':
            rtable = ReadOnlyListeningRedisRoutingTable(self.create_redis())
            rtable.start()
            atexit.register(rtable.stop, timeout=2)
            balancer = WeightedRoundRobinBalancer(rtable)
            return HostRouter(balancer)
        elif router_type == 'StaticRouter':
            host = self.env.get('galileo_router_static_host', 'http://localhost')
            return StaticRouter(host)
        elif router_type == 'DebugRouter':
            return DebugRouter()

        raise ValueError('Unknown router type %s' % router_type)

    def create_app_loader(self) -> AppClientLoader:
        loader = AppClientDirectoryLoader(self.env.get('galileo_apps_dir', os.path.abspath('./apps')))
        repo = RepositoryClient(self.env.get('galileo_apps_repository', 'http://localhost:5001'))

        return AppRepositoryFallbackLoader(loader, repo)

    def create_redis(self) -> redis.Redis:
        host = self.env.get('galileo_redis_host', 'localhost')

        if host.startswith('file://'):
            import redislite
            f_path = host.replace('file://', '')
            return redislite.Redis(dbfilename=f_path, decode_responses=True)

        params = {
            'host': host,
            'port': int(self.env.get('galileo_redis_port', '6379')),
            'password': self.env.get('galileo_redis_password', 'None'),
            'decode_responses': True,
        }
        logger.debug("establishing redis connection with params %s", params)

        return redis.Redis(**params)

    def create_exp_db(self) -> ExperimentDatabase:
        return create_experiment_database_from_env(self.env)


class DebugRouter(Router):

    def request(self, req: ServiceRequest) -> requests.Response:
        logger.debug('DebugRouter received service request %s', req)

        response = requests.Response()
        response.status_code = 200
        response.url = self._get_url(req)
        req.sent = time.time()

        return response

    def _get_url(self, req: ServiceRequest) -> str:
        return 'http://debughost' + req.path



import json
import logging
import os
import time
from contextlib import contextmanager
from json import JSONDecodeError
from typing import Callable

import pymq
import redis
from galileodb.model import Experiment, QueuedExperiment, ExperimentConfiguration, generate_experiment_id
from telemc import TelemetryRecorder

from galileo.controller import ExperimentController, RedisClusterController
from galileo.experiment import runner
from galileo.experiment.service import ExperimentService
from galileo.worker.api import StartTracingCommand, PauseTracingCommand

logger = logging.getLogger(__name__)


class ExperimentDaemon:
    POISON = '__POISON__'

    def __init__(self, rds: redis.Redis, create_recorder: Callable[[str], TelemetryRecorder],
                 exp_controller: ExperimentController, exp_service: ExperimentService) -> None:
        self.rds = rds
        self.create_recorder = create_recorder
        self.exp_controller = exp_controller
        self.exp_service = exp_service
        self.closed = False

        self.queue = None

    def close(self):
        self.closed = True
        if self.queue:
            self.queue.put_nowait(self.POISON)

    def run(self) -> None:
        self.queue = pymq.queue(ExperimentController.queue_key)

        try:
            logger.info('listening for incoming experiment...')
            while not self.closed:
                item = self.queue.get()

                if item == self.POISON:
                    break

                exp = None
                try:
                    logger.debug('got queued experiment %s', item)
                    exp, cfg = self.load_experiment(item)
                except Exception as e:
                    if exp and exp.id:
                        exp = self.exp_service.find(exp.id)
                        if exp is not None:
                            exp.status = 'FAILED'
                            self.exp_service.save(exp)

                    logger.exception('error while loading experiment from queue')
                    continue

                status = exp.status
                try:
                    pymq.publish(StartTracingCommand())
                    self.run_experiment(exp, cfg)
                    status = 'FINISHED'
                except Exception as e:
                    logger.error('error while running experiment: %s', e)
                    status = 'FAILED'
                finally:
                    pymq.publish(PauseTracingCommand())
                    logger.info("finalizing experiment %s", exp.id)
                    self.exp_service.finalize_experiment(exp, status)

        except KeyboardInterrupt:
            logger.info("interrupted while listening")

        logger.info('exiting experiment daemon loop')

    def run_experiment(self, exp: Experiment, cfg: ExperimentConfiguration):
        exp.status = 'IN_PROGRESS'
        exp.start = time.time()
        self.exp_service.save(exp)

        with managed_recorder(self.create_recorder, exp.id):
            logger.info("starting experiment %s", exp.id)
            runner.run_experiment(RedisClusterController(self.rds), cfg)

    def load_experiment(self, queued: QueuedExperiment) -> (Experiment, ExperimentConfiguration):
        exp = self.exp_service.find(queued.experiment.id) if queued.experiment.id else None

        if not exp:
            exp = queued.experiment

        if not exp.id:
            exp.id = generate_experiment_id()
        if not exp.name:
            exp.name = exp.id
        if not exp.creator:
            exp.creator = 'galileo-' + str(os.getpid())
        if not exp.created:
            exp.created = time.time()

        return exp, queued.configuration

    @staticmethod
    def _get_json_body(message):
        # parse JSON, mandatory attributes: id, name, creator, instructions
        try:
            return json.loads(message[1])
        except JSONDecodeError as e:
            logger.warning("JSON decoding error while parsing message body", e)
            return None


@contextmanager
def managed_recorder(create_recorder: Callable[[str], TelemetryRecorder], exp_id: str):
    r = create_recorder(exp_id)
    try:
        r.start()
        yield r
    finally:
        r.stop()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

tests_require = [
    'redislite>=5.0.124623',
    'timeout_decorator>=0.4.1',
    'pytest>=5.0.0',
    'pytest-cov>=2.7.1',
    'coverage>=4.5.3',
    'coveralls'
]
install_requires = [
    'galileo-db>=0.10.3',
    'telemc>=0.3.0',
    'requests>=2.20.1',
    'redis>=3.2.1',
    'pymq>=0.4.0',
    'pyyaml>=5.4.1',
    'click>=7.0',
]

setuptools.setup(
    name="edgerun-galileo",
    version="0.10.3.dev2",
    author="Thomas Rausch",
    author_email="t.rausch@dsg.tuwien.ac.at",
    description="Galileo: A framework for distributed load testing experiments",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/edgerun/galileo",
    packages=setuptools.find_packages(),
    test_suite="tests",
    tests_require=tests_require,
    install_requires=install_requires,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    entry_points={
        "console_scripts": ['galileo = galileo.cli.galileo:main']
    },
)



import argparse

import redis

from galileo.shell.shell import *


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--source', required=False, type=argparse.FileType('r'),
                        help='read and execute the given file instead of running in interactive mode')
    args = parser.parse_args()

    rds = redis.Redis(
        host=os.getenv('galileo_redis_host', 'localhost'),
        password=os.getenv('galileo_redis_password', None),
        port=int(os.getenv('galileo_redis_port', 6379)),
        decode_responses=True
    )
    init_module(rds, __name__)

    if args.source:
        exec(args.source.read(), globals())
        exit(0)

    if is_interactive:
        sys.ps1 = prompt

        print(banner)
        print('Welcome to the galileo shell!')
        print('')
        print('Type `usage` to list available functions')
        print('')
    else:
        sys.ps1 = ''


if __name__ == '__main__':
    # python -i shell.py will execute to here and then drop into the interactive shell
    main()
